#include "lusol.h"

int main(int argc, char *argv[]) {
/*----------------------------------- *
 *  Driver program for GPU L/U solve  *
 *  x = U^{-1} * L^{-1} * b           *
 *  Kernels provided:                 *
 *  CPU L/U solve                     *
 *  GPU L/U solve                     *
 *  GPU L/U solve w/ level-scheduling *
 *----------------------------------- */
  char *filename;
  int i,n,nnz,reorder,err; 
  REAL *h_b,*d_b,*d_x,*h_x0,*h_x1,*h_x2;
  struct coo_t h_coo;
  struct csr_t h_csr;
  struct lu_t h_lu, d_lu;
  struct level_t h_lev, d_lev;
  double e1,e2;
/*--------------------------------------*/
/*---------- cmd line arg */
  if (get_arg(argc,argv,&reorder,&filename))
    return 1;
/*--------------------------------------*/
  if (DOUBLEPRECISION)
    printf("L/U Solv DOUBLE precision\n");
  else
    printf("L/U Solv SINGLE precision\n");
/*---------- Init GPU */
  if (cuda_init() == -1) exit(-1);
/*---------- Read COO from mtx file */
  read_coo_MM(&h_coo, filename);
  n = h_coo.n;
  nnz = h_coo.nnz;
/*---------- COO -> CSR */
  COO2CSR(&h_coo, &h_csr);
/*------------ mmd reordering */
  if (reorder) {
    printf("  ** MMD reordering\n");
    int *perm, *iperm;
    Reorder_MMD(&h_csr, &perm, &iperm);
    RowColPermMat(&h_csr, perm);
    free(perm);
    free(iperm);
  }
/*----- sort each row by increasing col idx */
/*----- for the correctness of ILU(0) fact. */
  sortrow(n, h_csr.ia, h_csr.ja, h_csr.a);
/*------ ILU0 fact (on CPU) */
  if ((err=ilu0(&h_csr, &h_lu)) != -1) { 
    printf("ilu0 error:%d\n", err);
    exit(-1);
  }
  //printf("ilu0 ... Done\n");
/*------ Transfer LU from host to device */
  cudaMallocLU(n, h_lu.l.nnz, h_lu.u.nnz, &d_lu);
  LUHost2Device(&h_lu, &d_lu);
/*--------------------- vector b */
  srand(100);
  h_b = (REAL *) malloc(n*sizeof(REAL));
  for (i=0; i<n; i++)
    h_b[i] = rand() / (RAND_MAX + 1.0);
  d_b = NULL;
  cudaMalloc((void **)&d_b, (size_t)n*sizeof(REAL));
  cudaMemcpy(d_b, h_b, (size_t)n*sizeof(REAL),
             cudaMemcpyHostToDevice);
/*------------------------------------------------ */
/*------------- Start testing kernels ------------ */
/*------------- CPU L/U Sol */
  h_x0 = (REAL *) malloc(n*sizeof(REAL));
  luSolCPU(n, nnz, h_b, h_x0, h_lu.l.a, h_lu.l.ja, 
           h_lu.l.ia, h_lu.u.a, h_lu.u.ja, h_lu.u.ia);
/*------------- GPU L/U Solv w.o lev-sched */
  d_x = NULL;
  cudaMalloc((void **)&d_x, (size_t)n*sizeof(REAL));
  luSolv(n, nnz, &d_lu, d_x, d_b);
  h_x1 = (REAL *) malloc(n*sizeof(REAL));
  cudaMemcpy(h_x1, d_x, (size_t)n*sizeof(REAL),
             cudaMemcpyDeviceToHost);
  e1=error_norm(h_x0, h_x1, n);
  printf("  err norm %.2e\n", e1);
/*------------ GPU L/U Solv w/ Lev-Sched */
  cudaMemset(d_x, 0, n*sizeof(REAL));
  makeLevel(&h_lu, &h_lev, &d_lev);
  luSolvLev(n,nnz, &d_lu, &h_lev, &d_lev, d_x, d_b);
  h_x2 = (REAL *) malloc(n*sizeof(REAL));
  cudaMemcpy(h_x2, d_x, (size_t)n*sizeof(REAL),
             cudaMemcpyDeviceToHost);
  e2=error_norm(h_x0, h_x2, n);
  printf("  err norm %.2e\n", e2);
/*----------- Done free */
  free(h_b); free(h_x0); free(h_x1); free(h_x2);
  cudaFree(d_b); cudaFree(d_x);
  FreeCOO(&h_coo); FreeCSR(&h_csr);
  FreeLU(&h_lu); CudaFreeLU(&d_lu);
  FreeLev(&h_lev); CudaFreeLev(&d_lev);
/*---------- check error */
  cuda_check_err();
}
